#include "poi_processing.h"
#include <string.h>

#pragma warning (disable:4800)

using namespace FilterSpace;

bool POIProcessing(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	int NZ, ui;
	byte count_of_non_zero_neighbours[256];
	int  du[] = { -1,  0,  1, -1, 0, -1, 0, 1 };
	int  dv[] = { -1, -1, -1,  0, 0,  1, 1, 1 };

	for (int i = 0; i < 256; ++i) // 1-esek megszmllsa
	{
		for (NZ = 0, ui = i; ui; NZ++) ui &= ui - 1;
		count_of_non_zero_neighbours[i] = NZ;
	}

	int w          = f->width;
	int h          = f->height;
	int phase      = f->morph_phase;
	int foreground = f->morph_foreground;
	byte* img      = f->datain;
	byte* out      = f->dataout;
	int sin        = f->stridein;
	int sout       = f->strideout;

	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);
	int dx = sx + sw;
	int dy = sy + sh;
	int neighbours;
	int processed_pixels = 0;

	if (phase == 1)
	{
		for (int y = sy, yy = 0; y < dy; y++, yy++)
		{
			for (int x = sx, xx = 0; x < dx; x++, xx++)
			{
				byte val = img[sin * y + x];
				if (val == foreground) 
				{
					GET_NEIGHBOURS_BYTEPLANE(img, w, h, x, y, sin, foreground, neighbours);
					if (count_of_non_zero_neighbours[neighbours] > 2) val = 3;
				}
				out[sout * yy + xx] = val;
			}
			// FOR PERFORMANCE MEASUREMENT
			processed_pixels += sw;
			ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
			if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
		}
	}
	else if (phase == 2)
	{
		for (int y = sy, yy = 0; y < dy; y++, yy++)
		{
			for (int x = sx, xx = 0; x < dx; x++, xx++)
			{
				byte val = img[sin * y + x];
				if (val == foreground) 
				{
					GET_NEIGHBOURS_BYTEPLANE(img, w, h, x, y, sin, foreground, neighbours);
					if (count_of_non_zero_neighbours[neighbours] <= 1) val = 2;
				}
				out[sout * yy + xx] = val;
			}
			// FOR PERFORMANCE MEASUREMENT
			processed_pixels += sw;
			ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
			if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
		}
	}
	return true;
}


bool ExecutePOIProcessing(filtersettings* filset)
{ 
	bool b = false;
	if (filset)
	{
		filter_core_func f         = (::POIProcessing);
		int  iteration             = 0;
		int  eliminated_pixels     = 1;
		int  cpucores              = System::Environment::ProcessorCount;
		*(filset->processing_size) = filset->width * filset->height;
		filset->nCoreThreadId      = 0;
		filset->nCoreExecutionSize = 1; 

		*(filset->processing_ptr) = 0;
		b = ImageFilterBase::ExecuteFilterBox(filset, f);

		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}


