#include "polylinize.h"
#include "shp_polyline_export.h"
#include <list>
using namespace std;

#ifndef BETESZ_PONT
#define BETESZ_PONT(L,C,D) { if (D == 1) L.push_back(C); else L.push_front(C); }
#endif


bool get_neighbour(byte* img, int w, int h, int x, int y, int foreground, int& nx, int& ny, int which = 0)
{
	int cnt = 0;
	for (int my = -1; my <= 1; my++)
	{
		if (y + my < 0 || y + my > h - 1) continue;

		for (int mx = -1; mx <= 1; mx++)
		{
			if (x + mx < 0 || x + mx > w - 1) continue;

			if (img[(y + my) * w + x + mx] == foreground) 
			{
				if (cnt++ != which) continue;
				nx = x + mx;
				ny = y + my;
				return true;
			}
		}
	}
	return false;
}

bool search_pixel(byte* img, int w, int h, int& x, int& y, int foreground)
{ // kpes ott folytatni, ahol abbahagytuk, ezzel is hatkonyabb tve a keresst
	for (int my = y; my < h; my++)
	for (int mx = 0; mx < w; mx++)
	if (img[w * my + mx] == foreground) { x = mx; y = my; return true;}
	return false;
}


list<coord> process_segment(list<coord>& lst)
{
	list<coord> result;
	list<coord>::iterator it = lst.begin();

	int dx, dy, px, py, x, y, sx, sy;
	px = sx = (*it).x;
	py = sy = (*it).y;
	result.push_back(lst.front());
	if (lst.size() < 2) return result;
	it++;
	dx = (*it).x - px;
	dy = (*it).y - py;

	while (it != lst.end())
	{
		x = (*it).x;
		y = (*it).y;
		if (x - px != dx || y - py != dy) 
		{
			result.push_back(coord(px,py));
			sx = x;
			sy = y;
			dx = x - px;
			dy = y - py;
		}
		px = x;
		py = y;
		it++;
	}
	result.push_back(coord(px,py));
	return result;
}




void extract_line_segments(byte* img, int w, int h, const char* filename)
{
	int x, y, sx = 0, sy = 0;

	// Mikzben bejrjuk a zld pontokbl az utakat, eltroljuk, hogy 
	// mely piros pontokat rintettk. Ezeket utna letrljk.
	list<coord> processed_red_points;

	shp_polyline_exporter *spe = new shp_polyline_exporter(filename, w, h);

	// Zld pixelbl indulunk
	while (search_pixel(img, w, h, sx, sy, 2))
	{
		list<coord> line_segment;
		BETESZ_PONT(line_segment, coord(sx, h-sy), 1);
		REMOVE_POINT(img, w, sx, sy);

		for (int direction = 1; direction <= 2; direction++)
		{
			x = sx;
			y = sy;
			while (get_neighbour(img, w, h, x, y, 1, x, y)) // FEKETE pixelsorozat
			{
				BETESZ_PONT(line_segment,coord(x, h-y), direction);
				REMOVE_POINT(img, w, x, y);
			}
			if (get_neighbour(img, w, h, x, y, 2, x, y)) // ZLD pixel
			{
				BETESZ_PONT(line_segment,coord(x, h-y), direction);
				REMOVE_POINT(img, w, x, y);
				if (get_neighbour(img, w, h, x, y, 3, x, y)) // PIROS pixel
				{
					BETESZ_PONT(line_segment,coord(x, h-y), direction);
					processed_red_points.push_back(coord(x,y));
					if (get_neighbour(img, w, h, x, y, 4, x, y)) BETESZ_PONT(line_segment,coord(x, h-y), direction); // KK pixel
				}
				else if (get_neighbour(img, w, h, x, y, 4, x, y)) // KK pixel
				{
					BETESZ_PONT(line_segment,coord(x, h-y), direction);
				}
			}
			else if (get_neighbour(img, w, h, x, y, 3, x, y)) // PIROS pixel
			{
				BETESZ_PONT(line_segment,coord(x, h-y), direction);
				processed_red_points.push_back(coord(x,y));
				if (get_neighbour(img, w, h, x, y, 4, x, y)) BETESZ_PONT(line_segment,coord(x, h-y), direction); // KK pixel
			}
			else if (get_neighbour(img, w, h, x, y, 4, x, y)) // KK pixel
			{
				BETESZ_PONT(line_segment,coord(x, h-y), direction);
			}
		}

		line_segment = process_segment(line_segment);
		spe->write_polyline(line_segment);
	}

	for (list<coord>::iterator it = processed_red_points.begin(); it != processed_red_points.end(); it++)
		REMOVE_POINT(img, w, ((*it).x), ((*it).y));
	processed_red_points.clear();

	// Krt alkot fekete pixelek feldolgozsa 
	// (valdi kr, sehol sincs rajta elgazs vagy ms szn pont)
	sx = sy = 0;
	while (search_pixel(img, w, h, sx, sy, 1))
	{
		list<coord> line_segment;
		BETESZ_PONT(line_segment, coord(sx, h-sy), 1);
		REMOVE_POINT(img, w, sx, sy);

		// kr tbbi pontjnak feldolgozsa
		x = sx;
		y = sy;
		while (get_neighbour(img, w, h, x, y, 1, x, y)) // FEKETE pixelsorozat
		{
			BETESZ_PONT(line_segment,coord(x, h-y), 1);
			REMOVE_POINT(img, w, x, y);
		}

		BETESZ_PONT(line_segment, coord(sx, h-sy), 1);
		line_segment = process_segment(line_segment);
		spe->write_polyline(line_segment);
	}

	// Piros pixelbl indulunk
	sx = sy = 0;
	while (search_pixel(img, w, h, sx, sy, 3))
	{
		list<coord> line_segment;
		BETESZ_PONT(line_segment, coord(sx, h-sy), 1);
		REMOVE_POINT(img, w, sx, sy);
		int  nx, ny, neighbour_select = 0;
		bool first_is_blue = false;
		int  steps = 0;
		int  red_neighbour = 0;

		for (int direction = 1; direction <= 2; direction++)
		{
			x = sx;
			y = sy;

			while (true)
			{
				int prevx = x, prevy = y;
				if (get_neighbour(img, w, h, x, y, 4, x, y, neighbour_select)) // KK pixel
				{
					first_is_blue = (steps == 0 && direction == 1); 
					BETESZ_PONT(line_segment,coord(x, h-y), direction);
					nx = x;
					ny = y;
					neighbour_select++; // ha 2 kk szomszdja van, akkor azokat akarjuk sszektni
					break;
				}
				else if (get_neighbour(img, w, h, x, y, 3, x, y, red_neighbour)) // PIROS pixelsorozat
				{
					if (direction == 2 && (x-nx < 2 && x-nx > -2 && y-ny < 2 && y-ny > -2)) // nem lehet az 1. kk vg szomszdja
					{
						x = prevx; // maradunk ugyanott
						y = prevy;
						red_neighbour++; // msik piros szomszdot keresnk
						continue;
					}
					BETESZ_PONT(line_segment,coord(x, h-y), direction);
					REMOVE_POINT(img, w, x, y);
					neighbour_select = 0;
				}
				else
				{
					break; // valami hiba addott
				}
				steps++;
			}
		}		
		line_segment = process_segment(line_segment);
		spe->write_polyline(line_segment);
	}

	delete spe;
}

