#include "shp_polyline_export.h"
#include <string>
#include <sstream>
#pragma warning(disable:4267)
#pragma warning(disable:4996)
using namespace std;


shp_polyline_exporter::shp_polyline_exporter(const char *filename, int width, int height)
{
	w = width;
	h = height;
	rec_num = 0;
	fieldlen = 6;
	shp_file_header	shp_fh(0, width, height);

	std::string str = filename;
	str += ".shp";
	fp = fopen(str.c_str(), "wb");
	if (fp) fwrite(&shp_fh, sizeof(shp_file_header), 1, fp);
	
	// --------------------------

	str = filename;
	str += ".shx";
	fx = fopen(str.c_str(), "wb");
	if (fx) fwrite(&shp_fh, sizeof(shp_file_header), 1, fx);

	// --------------------------

	str = filename;
	str += ".dbf";
	fd = fopen(str.c_str(), "wb");

	unsigned char      data1;
	unsigned short int data2;
	unsigned int       data4;
	char               field[]   = "TYPE\0\0\0\0\0\0\0N\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";

	// file header
	fwrite(&(data1 = 3),               1,  1, fd);  // dBase5
	fwrite(&(data1 = 2010 - 1900),     1,  1, fd);  // year
	fwrite(&(data1 = 5),               1,  1, fd);  // month
	fwrite(&(data1 = 27),              1,  1, fd);  // day
	fwrite(&(data4 = rec_num),         4,  1, fd);  // number of records
	fwrite(&(data2 = 32 + 1 * 32 + 1), 2,  1, fd);  // header length
	fwrite(&(data2 = fieldlen),        2,  1, fd);  // record length
	fwrite(&(data4 = 0),               4,  1, fd);  // 0
	fwrite(&data4,                     4,  1, fd);  // 0
	fwrite(&data4,                     4,  1, fd);  // 0
	fwrite(&data4,                     4,  1, fd);  // 0
	fwrite(&data4,                     4,  1, fd);  // 0
	// field descriptor array
	field[16] = fieldlen - 1; // field length
	fwrite(field,                      1, 32, fd);  // field descriptor
	fwrite(&(data1 = 0x0D),            1,  1, fd);  // header terminator
}


shp_polyline_exporter::~shp_polyline_exporter()
{
	if (fp)
	{
		int fsize = ftell(fp) / 2;
		fsize = LE_TO_BE_INT(fsize);
		fseek(fp, 24, SEEK_SET);
		fwrite(&fsize, 4, 1, fp);
		fclose(fp);
	}
	if (fx)
	{
		int fsize = ftell(fx) / 2;
		fsize = LE_TO_BE_INT(fsize);
		fseek(fx, 24, SEEK_SET);
		fwrite(&fsize, 4, 1, fx);
		fclose(fx);
	}
	if (fd)
	{
		stringstream ss;
		string str;
		unsigned char data1;
		char content[] = "      ";
		int  fieldlen  = 6;
		// fields
		for (int i = 0; i < rec_num; i++)
		{
			ss << i + 1;
			ss >> str;
			ss.clear();
			fwrite(content, 1, 6 - str.length(), fd);
			fwrite(str.c_str(), 1, str.length(), fd);
		}
		// file terminator
		fwrite(&(data1 = 0x1A), 1, 1, fd);
		fclose(fd);
	}
}


void shp_polyline_exporter::write_polyline(std::list<coord> c)
{
	if (fp == 0) return;
	rec_num++;

	int tmp4;
	int count			= c.size();
	int offset			= ftell(fp) / 2;
	int content_length	= 24 + 8 * count;
	int record_number	= LE_TO_BE_INT(rec_num);
	offset				= LE_TO_BE_INT(offset);
	content_length		= LE_TO_BE_INT(content_length);
	
	// SHX --->
	fwrite(&offset, 4, 1, fx);
	fwrite(&content_length, 4, 1, fx);

	// SHP --->
	double box[4] = { w + 1.0, h + 1.0, -1.0, -1.0 };
	for (std::list<coord>::iterator it = c.begin(); it != c.end(); it++)
	{
		coord cc = *it;
		if (cc.x < box[0]) box[0] = cc.x;
		if (cc.y < box[1]) box[1] = cc.y;
		if (cc.x > box[2]) box[2] = cc.x;
		if (cc.y > box[3]) box[3] = cc.y;
	}

	// write record header
	fwrite(&record_number, 4, 1, fp);
	fwrite(&content_length, 4, 1, fp);
	// write record content
	fwrite(&(tmp4 = 3), 4, 1, fp);		// shape type
	fwrite(box, 8, 4, fp);				// box
	fwrite(&(tmp4 = 1), 4, 1, fp);		// num parts
	fwrite(&(tmp4 = count), 4, 1, fp);	// num points
	fwrite(&(tmp4 = 0), 4, 1, fp);		// parts 0 index
	
	for (std::list<coord>::iterator it = c.begin(); it != c.end(); it++) // points
	{
		coord cc = *it;
		box[0] = cc.x;
		box[1] = cc.y;
		fwrite(box, 8, 2, fp); // point
	}
}


