#include "thinning_regular.h"
#include <string.h>

#pragma warning (disable:4800)

using namespace FilterSpace;



void InitThinningRegular(filtersettings* f)
{
	unsigned int  i, j, ui, tr, Z0, NZ;

	// rowwise to counterclockwise.   BIT:  0->5, 1->4, 2->3, 3->6, 4->2, 5->7, 6->0, 7->1
	unsigned char rw_to_ccw_shift_to[8] = { 0x20, 0x10, 0x08, 0x40, 0x04, 0x80, 0x01, 0x02 };

	// i SORFOLYTONOS SZOMSZDOKAT REPREZENTL BJT BITJEIT 
	// rowwise_to_ccw_neigbours[i] RAJRSVAL ELLENTTES IRNYBAN FELSOROLVA BRZOLJA
	for (i = 0; i < 256; ++i) 
	{
		for (tr = j = 0, ui = i; ui; ui >>= 1, j++) if (ui & 1) tr |= rw_to_ccw_shift_to[j]; 
		f->rowwise_to_ccw_neigbours[i] = tr;
	}
	for (i = 0; i < 256; ++i) // 1-esek megszmllsa
	{
		for (NZ = 0, ui = f->rowwise_to_ccw_neigbours[i]; ui; NZ++) ui &= ui - 1;
		f->count_of_non_zero_neighbours[i] = NZ;
	}
	for (i = 0; i < 256; i++)
	{
		ui = f->rowwise_to_ccw_neigbours[i]; // row-wise -> counterclockwise ttrs
		ui = ui | ((ui & 1) << 8);        // az els s utols bit egyezzen meg
		for (Z0 = 0; ui; ui >>= 1) if ((ui & 1) == 0 && (ui & 2) == 2) Z0++; // 0->1 TMENETEK SZMLLSA
		f->zero_to_non_zero_transitions[i] = Z0;
	}
	// P2_P4_P6 multiply is zero <-> !(P2 = P4 = P6 = 1)
	// MASK P2, P4, P6, THEN COUNT NON-ZEROS. IF LESS THAN 3, MULTIPLY IS ZERO
	for (i = 0; i < 256; i++) f->P2_P4_P6_multiply_is_zero[i] = (f->count_of_non_zero_neighbours[i & 0x52] < 3) ? 1 : 0;
	for (i = 0; i < 256; i++) f->P2_P4_P8_multiply_is_zero[i] = (f->count_of_non_zero_neighbours[i & 0x58] < 3) ? 1 : 0;
}

/////////////////////////////
//       FILTER            //
/////////////////////////////

bool ThinningRegular(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// ITERATOR MAY PREVENT AN UNNEEDED THREAD TO RUN
	bool processing = (f->eliminated_pixels[f->nCoreThreadId] != -1);

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);

	int w = (sw) ? sw : f->width;
	int h = (sh) ? sh : f->height;
	int w_in = f->width;
	int h_in = f->height;


	unsigned char P1, P2, P4, NZ_P1, *p = f->dataout, *in = f->datain;
	int           x, y, xx, yy, cnt = 0, processed_pixels = 0, s = f->stridein;

	InitThinningRegular(f);

	for (y = 0, yy = sy; y < h; y++, yy++)
	{
		memcpy(p, in + yy * s + sx / 8, sw >> 3 );

		if (processing)
		{
			for (x = 0, xx = sx; x < w; x++, xx++)
			{
				// 0 = OBJEKTUM-PIXEL, 1 = HTTR-PIXEL
				if (GET_BITPIXEL_OF_ROW(in + yy * s,xx))		continue;
				GET_NEIGHBOURS_BITPLANE(in, xx, yy, w_in, h_in, s, P1);
				if (f->zero_to_non_zero_transitions[P1] != 1)	continue;
				NZ_P1 = f->count_of_non_zero_neighbours[P1];
				if (NZ_P1 < 2 || 6 < NZ_P1)						continue;
				GET_NEIGHBOURS_BITPLANE(in, xx, yy - 1, w_in, h_in, s, P2);
				if (f->zero_to_non_zero_transitions[P2] == 1 && f->P2_P4_P8_multiply_is_zero[P1] == 0) continue;
				GET_NEIGHBOURS_BITPLANE(in, xx - 1, yy, w_in, h_in, s, P4);
				if (f->zero_to_non_zero_transitions[P4] == 1 && f->P2_P4_P6_multiply_is_zero[P1] == 0) continue;
				SET_WHITEBITPIXEL_OF_ROW(p,x);
				cnt++;

			}
		}
		p += f->strideout;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;
		ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	f->eliminated_pixels[f->nCoreThreadId] = cnt;

	// FILTER CODE END -->

	return true;
}




bool ExecuteThinningRegular(filtersettings* filset)
{ 
	bool b = false;
	if (filset)
	{
		filter_core_func f         = (::ThinningRegular);
		int  iteration             = 0;
		int  eliminated_pixels     = 1;
		int  cpucores              = System::Environment::ProcessorCount;
		*(filset->processing_size) = filset->width * filset->height;
		filset->nCoreThreadId      = 0;
		filset->nCoreExecutionSize = 2; 
		filset->eliminated_pixels  = new int[cpucores];
		memset(filset->eliminated_pixels, 0, sizeof(int) * cpucores); 

		// A FILTER 1X MINDENKPPEN LEFUT. HA BEFEJEZI A FUTST, AKKOR AZ ELIMINLT PIXELEK SZMT 0-RA LLTJA

		memcpy(filset->dataout, filset->datain, filset->stridein * filset->height);
		filset->datain = new byte[filset->stridein * filset->height];

		while (eliminated_pixels && ((filset->filtering_iterations_limit == 0) || (iteration < filset->filtering_iterations_limit)))
		{
			*(filset->processing_ptr) = 0;

			memcpy(filset->datain, filset->dataout, filset->stridein * filset->height);

			b = ImageFilterBase::ExecuteFilterBox(filset, f);

			eliminated_pixels = 0; // VOLT MIT CSINLNI?
			for (int i = 0; i < cpucores; i++) 
			{
				if (filset->eliminated_pixels[i] == 0) filset->eliminated_pixels[i] = -1; // thread group not needed
				if (filset->eliminated_pixels[i]  > 0) 
				{
					eliminated_pixels += filset->eliminated_pixels[i];
					filset->eliminated_pixels[i] = 0;
				}
			}
			iteration++;
		}

		delete [] filset->datain;
		delete [] filset->eliminated_pixels;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}

